import processing.core.*; 
import processing.data.*; 
import processing.event.*; 
import processing.opengl.*; 

import java.util.HashMap; 
import java.util.ArrayList; 
import java.io.BufferedReader; 
import java.io.PrintWriter; 
import java.io.InputStream; 
import java.io.OutputStream; 
import java.io.IOException; 

public class KMeans extends PApplet {


/* 
 Processing sketch to illustrate how k-means clustering works in 2D.
 You can choose cluster centroids and the algorithm runs with cluster centroids 
 visually rearranging themselves
 */


/* ************* PARAMETERS ************* */

int NUM_POINTS = 200; // Number of data points to cluster
boolean CLUSTERED = true; // Whether the random data are slightly clustered or not

boolean OUTPUT_IMAGE = false; // Whether or not to write an image at each iteration

/* ************* ************* */

int imageNumber = 0; // used when writing images for filenames
String imageType = ".jpeg";

// Arrays to store the x,y and z coordinates of each data point
ArrayList<Float> dataX = new ArrayList<Float>(NUM_POINTS); // x coord
ArrayList<Float> dataY = new ArrayList<Float>(NUM_POINTS); // y coord 
ArrayList<Float> dataZ = new ArrayList<Float>(NUM_POINTS); // z coord
ArrayList<Integer> colours = new ArrayList<Integer>(NUM_POINTS); // colour

// All the cluster centres (Centre is an inner-class)
ArrayList<Centre> centres = new ArrayList<Centre>();

// Keep track of where in the algorithm we are:
//  0 - beginning
//  1 - have re-coloured data
//  2 - have moved centres
int step = 0;

public void setup() {

  size(800, 600, P2D);

  // Generate some points
  if (CLUSTERED) {
    makeClusteredData(); // Data is slightly clustered
  }
  else {
    makeRandomData();
  }  

  // Draw the points
  drawData();
}

public void draw() {

  // Draw data and cluster centres
  //drawData();
}

/* Iterate the alrogithm */

public void step() {

  imageNumber++;

  if (this.centres.isEmpty()) {
    println("No cluster centres, cannot step algorithm");
    return;
  }

  if (step == 0 || step == 2) { // At beginning or have just moved centres
    // Reset all centres, they have just moved so will now will have different points
    for (Centre c:centres) {
      c.data=new ArrayList<Integer>();
    }
    // Assign each point to its nearest cluster centre
    float d, min_d; // distance
    Centre centre; // Closest centre
    for (int i=0; i<NUM_POINTS; i++) {
      min_d = 99999;
      centre = null;
      for (Centre c:centres) {
        d = c.distance(dataX.get(i), dataY.get(i), dataZ.get(i));
        if (d<min_d) {
          min_d = d;
          centre=c;
        }
      } // for Centres
      assert centre!= null; // chech have found a closest one
      colours.set(i, centre.c); // Re-colour the data point using cluster centre
      centre.data.add(i); // Tell the centre that this datapoint is in its catchment
    }
    step = 1; // Indicate where in the algorithm we are
  } // if step

  else if (step==1) { // Have just coloured data, now move centres
    for (Centre c:centres) {
      if (c.data==null) { // this can happen if a new centre has just been added and no data calclated yet
        continue;
      }
      float[] xy = c.findCentrePoint();      
      c.x = xy[0];
      c.y = xy[1];
    }

    step = 2; // So we know that next time we need to re-colour the data
  }

  //redraw(); // Indicate a redraw is needed

  drawData(); // redraw everything
} // step()



/* When the mouse is clicked create a new cluster centre */
public void mouseClicked() {

  println("Mouse clicked: "+mouseX+","+mouseY);

  // Add a new centre at the clicked point (random color)
  int c = color(PApplet.parseInt(random(0.0f, 255)), PApplet.parseInt(random(0.0f, 255)), PApplet.parseInt(random(0.0f, 255)));
  centres.add(new Centre(mouseX, mouseY, c));
  fill(c);
  ellipse(mouseX, mouseY, 20, 20); // x,y,width,height
  if (OUTPUT_IMAGE) {
    imageNumber++;
    save("out"+imageNumber+imageType);
  }
}

public void keyPressed() {

  println("Received key '"+key+"'");
  if (key==' ') { // space
    step();
  }
  else if (key=='r' || key=='r') {    
    // reset
    dataX = new ArrayList<Float>(NUM_POINTS);
    dataY = new ArrayList<Float>(NUM_POINTS);
    dataZ = new ArrayList<Float>(NUM_POINTS);
    colours = new ArrayList<Integer>(NUM_POINTS);
    centres = new ArrayList<Centre>();
    int step = 0;
    if (CLUSTERED) {
      makeClusteredData(); // Data is slightly clustered
    }
    else {
      makeRandomData();
    }
    drawData();
  }
}

public void drawData() {
  background(100);
  // Draw data
  for (int i=0; i<NUM_POINTS; i++) {
    pushMatrix();
    fill((int)colours.get(i));
    translate(dataX.get(i), dataY.get(i));
    //fill(color(dataX.get(i),dataY.get(i),0));
    ellipse(0, 0, 10, 10); // x,y,width,height
    popMatrix();
  }

  // Draw cluster centres
  for (Centre c:centres) {
    fill(c.c); // Colour of the centre
    ellipse(c.x, c.y, 20, 20);
  } // for centres

  if (OUTPUT_IMAGE) {
    save("out"+imageNumber+imageType);
  }
} // drawData



/* Create some points in random places. */
public void makeRandomData() {

  for (int i=0; i<NUM_POINTS; i++) {

    dataX.add(random(0, width));
    dataY.add(random(0, height));
    dataZ.add(0.0f); // No z coord in 2D mode

    // Default colour for points is grey
    colours.add(color(100, 100, 100));
  }
}

/* Create some points in random places, clustered around 4 points areas */
public void makeClusteredData() {

  // Long winded way to create four points for the data to cluster around (centreX and centreY variables):  
  int[] clustersX = new int[] {
    width/4, (width*3)/4
  };
  int[] clustersY = new int[] {
    height/4, (height*3)/4
  };
  float clusterX=0.0f;
  float clusterY=0.0f; 
  int numPoints = NUM_POINTS/4;
  for (int c=0; c<4; c++) {
    switch (c) {
    case 0: 
      clusterX = clustersX[0];
      clusterY = clustersY[0];
      break;
    case 1:
      clusterX = clustersX[0];
      clusterY = clustersY[1];
      break;
    case 2:
      clusterX = clustersX[1];
      clusterY = clustersY[1];
      break;
    case 3:
      clusterX = clustersX[1];
      clusterY = clustersY[0];
      break;
    } // switch

    // Now create points around each centroid:
    for (int i=0; i<numPoints; i++) {

      float x = randomNormal(clusterX, clusterX+width*0.12f);
      float y = randomNormal(clusterY, clusterY+height*0.12f);

      dataX.add(x);
      dataY.add(y);
      dataZ.add(0.0f); // No z coord in 2D mode

      // Default colour for points is grey
      colours.add(color(100, 100, 100));
    } // for numPoints
  } // for clusters
} // makeClusteredData


/* randomNormal function adapted from: http://www.openprocessing.org/sketch/1254 */
// return a random value in the range [min..max] with normal distribution around min.
// Implements the Marsaglia Polar Method, as described in wikipedia, but only returns one of the values.
public float randomNormal(float min, float max)
{
  // Calcuate random from 0 - 1 around 0 
  float x = 1.0f, y = 1.0f, 
  s = 2.0f; // s = x^2 + y^2
  while (s >= 1.0f)
  {
    x = random(-1.0f, 1.0f);
    y = random(-1.0f, 1.0f);
    s = x*x + y*y;
  }


  float value = x * sqrt(-2.0f * log(s)/s);

  // Scale value to range [min..max] and return

  return value*(max-min) + min;
}

/* Class for representing cluster centres (either 2D or 3D) */
class Centre {

  float x;
  float y;
  float z;
  int c;
  ArrayList<Integer> data = null; // The indices of the data points associated with this centre 

  Centre(float x, float y, int c) {
    this.x=x;
    this.y=y;
    this.c=c;
  }

  Centre(float x, float y, float z, int c) {
    this(x, y, c);
    this.z = z;
  }

  /* The distance between this cluster Centre and the given coordinates (3D) */
  public float distance(float x, float y, float z) {
    return dist(x, y, z, this.x, this.y, this.z);
  }

  /* The distance between this cluster Centre and the given coordinates (2D) */
  public float distance(float x, float y) {
    return dist(x, y, this.x, this.y);
  }

  /* Find the centre point of this Centre's data. Returns float array with x,y,(z) coords */
  public float[] findCentrePoint() {
    // Use average x,y at the moment
    float totx = 0, toty=0;
    for (Integer index:this.data) { // Iterate over indices of data in this Centre's catchment
      totx += dataX.get(index);
      toty += dataY.get(index);
    }
    return new float[] {
      totx/this.data.size(), toty/this.data.size()
      };
    }
  }

  static public void main(String[] passedArgs) {
    String[] appletArgs = new String[] { "KMeans" };
    if (passedArgs != null) {
      PApplet.main(concat(appletArgs, passedArgs));
    } else {
      PApplet.main(appletArgs);
    }
  }
}
